dict_str = as-with dic def dic -> map(f, _) ->> ' '.join where:
        f = .key -> equal where:
            equal = f'{key} = "{dic[key]}"'
    
default_attr = . attr, value -> .func -> _f where:
    def _f(*args, **attributes):
        if attr not in attributes:
            attributes[attr] = value
        elif value not in attributes[attr]:
            attributes[attr] = f"{value} " + attributes[attr] 
        return func(*args, **attributes)

attrset_sugar = as-with conf, attributes def as attr, value def None where:
    if attr in attributes: 
        conf[attr] = attributes[attr]
        del attributes[attr]
    else:
        conf[attr] = value
    


dict_init_key = . dic -> . attr, type -> None where:
        if attr not in dic:
            dic[attr] = type()
            
            
JS_template =\
"""
{indent}<script>
{indent}{Indent_unit}{JS}
{indent}</scirpt>
"""

class ANY:
    def __eq__(self,v):
        return True
Any = ANY()

class Rule:
    def __init__(self, f):
        self.f = f
    def __eq__(self, v):
        if self.f(v):
            return True
        else:
            return False
        
        
        
        

    


from . import foreach, andThen, compose, fastmap
from .utils import dict_str, default_attr,dict_init_key, JS_template, attrset_sugar
from collections import defaultdict
from jinja2 import Template
from copy import deepcopy

"""Some default configurations are defined here."""
default_conf=dict(Indent_unit = "    ") 
get_type  = . obj  ->  obj.__class__
checktype = . type -> .obj -> get_type(obj) -> issubclass(_, type)
def gen_helper(render) -> ('dict -> dict', 'otherwise -> str'):
    """
    Recursively render the objects.
    """
    condef render:
        
        (get_type)
        case dict               =>
            render ->> foreach( key_func ) where:
                key_func = as-with key def None where:
                    render[key] = gen_helper(render[key]) -> _ if key != 'attributes_dict' else dict_str(_)
                    
        (checktype(abstract_object))
        case True               => 
            render = render.gen()
      
                    
        (get_type)
        case Seq                =>
            render = render ->> fastmap(gen_helper) => tuple => '\n'.join
            
        otherwise               =>
                        pass
    return render
                
""" Meta """
                
class abstract_object:
    """
    Everything in Materialize-CSS has been abstracted to this Class.
    Everything in Materialize-CSS has 2 methods in common:
        1. setIndent :: int -> None
            If inherit the Class  "indent_setter"
            
                Recursively change the indentation of the object with the following rules:
                    <subobject's indentation>  = (<indentation> + 2) * Indent_unit
                    "Indent_unit" is defined in "incantation.Module.abst.default_conf"
                    
            Else 
                Do Nothing
                
        2. gen :: () -> str
            Generate the html codes from the object.
    """
    def __init__(self,*args,**kwargs):
        self.conf = dict()
        self.body = ""
        eval("self.init(*args,**kwargs)")
        self.common_init()
        self.JS = ""
        

    def setIndent(self, i:int):
        """
        Default to do nothing for some items without indentation. 
        Inherit the Class "indent_setter" to set Indent recursively.
        PS: Inherit the Class "indent_setter" before inheriting "abstract_object".
        """
        return self
    
    def gen(self):
        """
        Generate the html codes from the object.
        """
        render   = deepcopy(self.conf)
        rendered =  eval("Template(self.body).render(**gen_helper(render) )")
        if self.JS:
            rendered = rendered + JS_template.format(JS = (self.conf['indent']+default_conf['Indent_unit']).join(self.JS.splitlines()), 
                                                     indent =self.conf['indent'], 
                                                     Indent_unit =default_conf['Indent_unit'])
        return rendered
        
    def common_init(self):
        """
        Do some common initial actions when initializing the Materialize-CSS object.
        """
        self.conf.update(default_conf)
    
    def cons_class(self, string:str):
        dict_init_key(self.conf['attributes_dict'])('class', str)
        self.conf['attributes_dict']['class'] = f"{string} {self.conf['attributes_dict']['class']}"
        return self
    
    def append_class(self, string:str):
        dict_init_key(self.conf['attributes_dict'])('class', str)
        self.conf['attributes_dict']['class'] += f" {string}"
        return self
    
    def cons_attr(self, attr:str):
        dict_init_key(self.conf['attributes_dict'])(attr, str)
        def _f(string:str):
            self.conf['attributes_dict'][attr] = f"{string} {self.conf['attributes_dict'][attr]}"
            return self
        return _f
    
    def append_attr(self, attr:str):
        dict_init_key(self.conf['attributes_dict'])(attr, str)
        def _f(string:str):
            self.conf['attributes_dict'][attr] += f" {string}"
            return self
        return _f
    
        


""" Trait """

class indent_setter:
    """
    A trait of the objects in Materialize-CSS.
    To set the indentations of objects recursively.
    """
    def setIndent(self, i:int):
        self.conf["indent"] = i*default_conf["Indent_unit"] 
        self.conf ->> foreach(job) where:
            job = as-with conf_key def None where:
                if   self.conf[conf_key] -> isinstance(_,  indent_setter):
                     _.setIndent(i+2)
                elif self.conf[conf_key] -> isinstance(_,  Seq):
                    foreach(for_each_item)(_) where:
                        for_each_item = . each_item ->  None where:
                            if isinstance(each_item, indent_setter):
                                each_item.setIndent(i+2)
        return self
                    
class Seq(list):
    """
    A kind of container of the objects in Materialize-CSS.
    This trait means that an object can have multiple items to be rendered, for instance:
        * form
        * table
        * row
    For example:
         cols = Seq(
                col(content1, dict(m=6,s=12,l=12)), 
                col(content2, dict(m=6,s=12,l=12)),
                col(content3, dict(m=6,s=12,l=12))
            )
         Row = row(cols)
    """
    def __init__(self, *args):
        return super(Seq, self).__init__(args)
    

                    
                    





from ..abst import abstract_object,indent_setter, Seq, default_attr
from ..utils import Rule
from ..abst import Template
class container(indent_setter, abstract_object):
    """
    See http://materializecss.com/grid.html.
    user help : >> help (container.init)
    ------------------------------
        Guide:
            part1 = container(content) where:
                content = r where:
                    r   = row(Seq(col1, col2)) where:
                        col1 = col("nothing")
                        col2 = col("nothing")
    """
    
    @default_attr(attr = "class", value = "container")
    def init(self, content = "" , **attributes):
        body   = \
"""
{{indent}}<div {{attributes_dict}}>
{{indent}}<!-- Page Content goes here -->
{{indent+Indent_unit*1}}{{content}}
{{indent}}</div>
"""            
        self.conf.update(dict(content = content, indent = '',  attributes_dict = attributes))
        self.body = body

    def contains(self, content):
        self.conf["content"] = content
        return self
        
    
class row(indent_setter, abstract_object):
    """
    See http://materializecss.com/grid.html.
    user help : >> help (grid.init)
    ------------------------------
        Guide:
            r = row(Seq(col1('content1', grid(s = 12)),
                    col2('content2', grid(s = 12)),
                    col3('content3', grid(s = 12)),
                    ))
            r.append_class("center")
            
    """
    
    @default_attr(attr = "class", value = "row")
    def init(self, content : Seq , **attributes):
        body = \
"""
{{indent}}<div {{attributes_dict}}>
{{indent+Indent_unit}}{{content}}
{{indent}}</div>
"""
        self.conf.update(dict(content = content, indent = " ", attributes_dict = attributes))
        self.body = body

class col(indent_setter, abstract_object):
    """
    See http://materializecss.com/grid.html.
    user help : >> help (col.init)
    ------------------------------
        Guide:
                >> col("some text", grid(m = 6, s= 6, l = 12))
    """
    def init(self, content, grid : "CSS.Grid.grid" ,  **attributes):
        body =\
"""
{{indent}}<div {{attributes_dict}}>
{{indent+Indent_unit}}{{content}}
{{indent}}</div>
"""
        @default_attr(attr = "class", value = grid.gen())
        def _init_(content, **attributes):    
            self.conf.update(dict(content = content, indent = " ", attributes_dict = attributes))
            self.body = body
        _init_(content, **attributes)
    
class divider(indent_setter, abstract_object):
    """
    See http://materializecss.com/grid.html.
        Guide :
            a_divider = divider() # take no parameter!
            
    """
    def init(self,*args, **kwargs):
        self.body =\
"""
{{indent}}<div class="divider"></div>
"""
        self.conf.update(dict(indent = " "))
        
class section(indent_setter, abstract_object) :
    """
    See http://materializecss.com/grid.html.
    ------------------------------
    Guide :
        >>> sec1 = section("some text", **{'class':'center'})
        >>> inst_container : Module.CSS.Grid.container
        >>> sec2 = section(inst_container, someattr = ...)
    """
    @default_attr(attr = "class", value = "section")
    def init(self,content, **attributes):
        body =\
"""
{{indent}}<div {{attributes_dict}}>
{{indent+Indent_unit}}{{content}}
{{indent}}</div>
"""
        self.conf.update(dict(content = content, indent = '',  attributes_dict = attributes))
        self.body = body
        

class grid(abstract_object, dict):
    """
    grid(m = 1, s = 2 ,l =3).push
    default:
        s:m:l = 12 : 6 : 4
    
    user help:
        >>> col("content here!", grid(m=5, s= 8, l=3).roffset(grid(s=3,m=1,l=1)) )
        >>> col("content here!", grid(m=5, s= 8, l=3).roffset(grid(s=3,m=1,l=1)).loffser(3) ) # s=3, m = 3//2, l = 3//3
        >>> col("content here!", grid(m=5, s= 8, l=3)          \
                                    .roffset(grid(s=3,m=1,l=1))\
                                    .loffser(s = 3, m = 2, l =1)
                )
    """    
    def init(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)       
        self.update(dict(loffset = (), roffset=() ))
        
        s = None if 's' not in self else self['s']
        m = None if 'm' not in self else self['m']
        l = None if 'l' not in self else self['l']
        condef[] (s,m,l):
            
            case (None, None, None) =>
                raise ValueError("Number of input arguments cannot be zero!!! ")
            
            case (Rule(.x->x.__class__ is int), None, None) =>
                self['m'] = self['s']//2
                self['l'] = self['s']//3
            
            case (None, Rule(.x->x.__class__ is int), None) =>
                self['s'] = self['m']*2
                self['l'] = self['s']//3
            
            case (None, None, Rule(.x->x.__class__ is int)) =>
                self['s'] = self['l']*3
                self['m'] = self['s']//2
                
            (.x -> map(type, x) ->> tuple)
            case (int,int,int)                              =>
                pass
            
            otherwise                                       =>
                raise ValueError("Do not support initializing <class 'grid'> with these arguments!!!")
            
    
    def loffset(self, *single, **dic):
        tup = single[0] if single else dic
        condef tup:
            (type)
            case int =>
                self['loffset'] = dict(s = tup, m = tup//2, l = tup//3)
                
            (.x -> map(type, x.values()) ->> tuple)
            case (int,int,int) =>
                self['loffset'] = tup
        
            otherwise          =>
                raise ValueError("Do not support setting left-offset with these arguments!!!")
        return self
    def push(self, *single, **dic):
        """ equals `grid.loffset`"""
        return self.loffset( *single, **dic)    
                
    def roffset(self, *single, **dic):
        tup = single[0] if single else dic
        condef tup:
            (type)
            case int =>
                self['roffset'] = dict(s = tup, m = tup//2, l = tup//3)
                
            (.x -> map(type, x.values()) ->> tuple )
            case (int,int,int) =>
                self['roffset'] = tup
        
            otherwise          =>
                raise ValueError("Do not support setting right-offset with these arguments!!!")
        return self
    def pull(self, *single, **dic): 
        """ equals `grid.roffset`"""
        return self.roffset(*single, **dic)
    
    def gen(self):
        ret = Template("col s{{s}} m{{m}} l{{l}}").render(**self)
        if self['loffset']:
            ret += Template(" push-s{{s}} push-m{{m}} push-l{{l}}").render(**self['loffset'])
        if self['roffset']:
            ret += Template(" pull-s{{s}} pull-m{{m}} pull-l{{l}}").render(**self['roffset'])
        return ret
        
            

        
                
            
            
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 26 16:14:34 2017

@author: misakawa
"""
from ..abst import abstract_object,indent_setter
class table(indent_setter, abstract_object):
    """
    See http://materializecss.com/table.html
    user help : >> help (table.init)
        Guide:
            >> t = table( 
                    ['name','email','GPA'],
                    [
                            ['thautwarm','twshere', '3.7'],
                            ['nightynight','thaut','4.0']
                    ]
                        )
            >> t
            <table >
                <thead>
                    <tr>
                    
                        <th>name</th>
                    
                        <th>email</th>
                    
                        <th>GPA</th>
                    
                    </tr>
                </thead>
                <tbody>
                    
                    <tr>
                        
                        <td>thautwarm</td>
                        
                        <td>twshere</td>
                        
                        <td>3.7</td>
                        
                    </tr>
                    
                    <tr>
                        
                        <td>nightynight</td>
                        
                        <td>thaut</td>
                        
                        <td>4.0</td>
                        
                    </tr>
                    
                </tbody>
            </table>
    """
    
    def init(self, column_names : list, data : list,   **attributes):

        body = \
"""
{{indent}}<table {{attributes_dict}}>
{{indent+Indent_unit}}<thead>
{{indent+Indent_unit*2}}<tr>
{{indent+Indent_unit*2}}{% for column_name in column_names%}
{{indent+Indent_unit*3}}<th>{{column_name}}</th>
{{indent+Indent_unit*2}}{% endfor %}
{{indent+Indent_unit*2}}</tr>
{{indent+Indent_unit}}</thead>
{{indent+Indent_unit}}<tbody>
{{indent+Indent_unit*2}}{% for item in data%}
{{indent+Indent_unit*2}}<tr>
{{indent+Indent_unit*3}}{% for cell in item%}
{{indent+Indent_unit*3}}<td>{{cell}}</td>
{{indent+Indent_unit*3}}{% endfor %}
{{indent+Indent_unit*2}}</tr>
{{indent+Indent_unit*2}}{% endfor %}
{{indent+Indent_unit}}</tbody>
{{indent}}</table>
"""
        self.conf.update(indent = " ", data = data, column_names = column_names, attributes_dict = attributes)
        self.body = body
    
        


# -*- coding: utf-8 -*-
from ..abst import abstract_object,indent_setter, default_attr

class img(indent_setter, abstract_object):
    """
    See http://materializecss.com/media-css.html
    user help : >> help (img.init)
    ------------------------------
        Guide:
            >> im = img(src = "cool_pic.jpg")
            >> im.append_class("responsive-img")
            >> im.append_class("class")
        which equals:
            >> im = img(**{'src': 'cool_pic.jpg', 'class':'responsive-img class'})
    """
    def init(self, **attributes):
        body = \
"""
{{indent}}<img  {{attributes_dict}}>
"""
        self.conf.update(dict(indent = " ", attributes_dict=attributes))
        self.body = body
        

class video_container(indent_setter, abstract_object):
    """
    See http://materializecss.com/media-css.html
    user help : >> help (video_container.init)
    ------------------------------
        Guide:
            >> video = video_container(src = "//www.youtube.com/embed/Q8TXgCzxEnw?rel=0")
            
        Take care that `video_container` takes some default arguments:
            
            video = video_container(src = "//www.youtube.com/embed/Q8TXgCzxEnw?rel=0")
        equals 
            video = video_container(src   = "//www.youtube.com/embed/Q8TXgCzxEnw?rel=0",
                                    width = 853,
                                    height= 480,
                                    frameborder = 0)
    """
    @default_attr(attr = 'frameborder', value = "0")
    @default_attr(attr = 'width' , value = "853")
    @default_attr(attr = 'height', value = "480")
    def init(self, **attributes):
        body = \
"""
{{indent}}<div class="video-container">
{{indent+Indent_unit}}<iframe {{attributes_dict}} allowfullscreen></iframe>
{{indent}}</div>
"""
        self.conf.update(indent = " ", attributes_dict = attributes)
        self.body = body
        
class responsive_video(indent_setter, abstract_object):
    """
    See http://materializecss.com/media-css.html
    user help : >> help (video_container.init)
    ------------------------------
        Guide:
            >> video = responsive_video(src = "movie.mp4")
        which equals:
            >> video = responsive_video(src = "movie.mp4", type = "video/mp4")
    """
    
    @default_attr(attr = 'type', value = 'video/mp4')
    def init(self, **attributes):
        body = \
"""
{{indent}}<video class="responsive-video" controls>
{{indent+Indent_unit}}<source {{attributes_dict}}>
{{indent}}</video>
"""
        self.conf.update(indent = " ", attributes_dict = attributes)
        self.body = body

    
        
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 31 22:42:38 2017

@author: misakawa
"""


from ..abst import abstract_object,indent_setter, default_attr, attrset_sugar, Seq

class navbar(indent_setter, abstract_object):
    """
    See http://materializecss.com/navbar.html
    user help : >> help (navbar.init)
        Guide:
            navbar(
                   [dict(href = 'https://baidu.com',  name = 'link1'),
                    dict(href = 'https://google.com', name = 'link2')
                   ],
                   href = 'https://github.com/thautwarm', 
                   name = 'Logo'
                   )
    """
    def init(self, content:(Seq,'or',list,'or', tuple), **attributes):
        sugar = attrset_sugar(self.conf, attributes)
        sugar('href', '!#')
        sugar('loc', 'left')
        sugar('name', 'Logo')
        sugar('ul_id', 'nav-mobile')
        self.conf['loc'] = ('left', 'right') if self.conf['loc']  == 'left' else ('right', 'left')
        
        condef content:
            {.obj, typefamily -> type(obj) in typefamily}
            case {list,tuple} =>
                body = \
"""
{{indent}}<nav>
{{indent+Indent_unit}}<div {{attributes_dict}}>
{{indent+Indent_unit}}<a href="{{href}}" class="brand-logo {{loc[0]}}">{{name}}</a>
{{indent+Indent_unit}}<ul id="{{ul_id}}" class="{{loc[1]}} hide-on-med-and-down">
{{indent+Indent_unit}}{% for item in content %}
{{indent+Indent_unit*2}}<li><a href="{{item.href}}">{{item.name}}</a></li>
{{indent+Indent_unit}}{% endfor %}
{{indent+Indent_unit}}</ul>
{{indent+Indent_unit}}</div>
{{indent}}</nav>
"""
            (.x->isinstance(x, abstract_object))
            case True =>
                body = \
"""
{{indent}}<nav>
{{indent+Indent_unit}}<div {{attributes_dict}}>
{{indent+Indent_unit}}<a href="{{href}}" class="brand-logo {{loc[0]}}">{{name}}</a>
{{indent+Indent_unit}}<ul id="{{ul_id}}" class="{{loc[1]}} hide-on-med-and-down">
{{indent+Indent_unit*2}}{{content}}
{{indent+Indent_unit}}</ul>
{{indent+Indent_unit}}</div>
{{indent}}</nav>
"""
        @default_attr(attr = 'class', value = 'nav-wrapper')
        def _init_(content, **attributes):
            self.conf.update(dict(content = content, indent = " ", attributes_dict = attributes))
            self.body = body
        _init_(content, **attributes)
        
        



from ..abst import abstract_object

class color(abstract_object):
        """
        Number of input arguments should be 1 or 3.
        See http://materializecss.com/color.html.
        """
        def init(self,  major  : "blue,red,yellow, ..."     = "blue",
                        degree : "lighten, darken, accent"  = "",
                        micro  : "int, None" = ""):
            if degree and micro:
                self.conf.update(dict(major = major , degree = degree, micro = micro))
                self.body = "{{major}} {{degree}}-{{micro}}"
            else:
                self.conf.update(dict(major = major))
                self.body = "{{major}}"

            
Red    = color( major = 'red'   )
Pink   = color( major = 'pink'  )
Purple = color( major = 'purple')
Indigo = color( major = 'indigo')
Blue   = color( major = 'blue'  )
Cyan   = color( major = 'cyan'  )
Teal   = color( major = 'teal'  )
Green  = color( major = 'green' )
Lime   = color( major = 'lime'  )
Yellow = color( major = 'yellow')
Amber  = color( major = 'amber' )
Orange = color( major = 'orange')
Brown  = color( major = 'brown' )

BlueGrey   = color( major = 'blue-grey'    )
LightGreen = color( major = 'light-green'  )
LightBlue  = color( major = 'light-blue'   )
DeepPurple = color( major = ' deep-purple' )
       
  
    
    



    

